local gfx, timer, fs = love.graphics, love.timer.getTime, love.filesystem
local font = gfx.setNewFont('pf_tempesta_seven_bold.ttf', 8)

local BENCHMARK_TIME = 1 -- approx. number of seconds each benchmark runs

local testData = {
	sequentialManyNumbers = function()
		local result = {}
		for i = 1, 2 ^ 16 do
			result[i] = i
		end
		return result
	end,

	sequentialManyU32 = function()
		local result = {}
		for i = 1, 2 ^ 16 do
			result[i] = i
		end
		return result
	end,

	sequentialFewNumbers = function()
		local result = {}
		for i = 1, 32 do
			result[i] = i
		end
		return result
	end,

	sequentialFewU8 = function()
		local result = {}
		for i = 1, 32 do
			result[i] = i
		end
		return result
	end,

	deepTable = function()
		local result = {
			number = 42,
			string = 'text',
			bool = true,
			nested = { }
		}

		current = result.nested
		for i = 1, 1000 do
			current.nested = {
				number = i,
				string = 'text ' .. i,
				bool = true,
			}
			current = current.nested
		end

		return result
	end,

}

local function generateData()
	local data = {}
	for name, func in pairs(testData) do
		data[name] = func()
	end
	return data
end

local benchmarks = {
	libraries = {}
}

local function loadBenchmarks()
	local dir = fs.getDirectoryItems('benchmarks')
	for i = 1, #dir do
		local name = dir[i]:gsub('%.lua', '')
		local benchmark = require('benchmarks.' .. name)

		local lib = {
			description = benchmark.description,
			serialize = {},
			deserialize = {}
		}

		for k, v in pairs(benchmark.serialize) do
			lib.serialize[k] = v
		end
		for k, v in pairs(benchmark.deserialize) do
			lib.deserialize[k] = v
		end
		lib.results = {}

		benchmarks.libraries[name] = lib
	end
end

local benchmark
local frame = 0

local function run(benchmark, data)
	local time, counter, resultData = 0, 0
	local okay, result = pcall(function()
		local start = timer()
		repeat
			resultData = benchmark(data)
			time = timer() - start
			counter = counter + 1
		until time >= BENCHMARK_TIME

		return {
			count = counter,
			time = time,
			data = resultData,
		}
	end)

	if not okay then
		result = {
			count = 0,
			time = BENCHMARK_TIME,
			error = result
		}
	end

	return result
end

local function runBenchmarks()
	local testData = generateData()
	for libName, lib in pairs(benchmarks.libraries) do
		for test, data in pairs(testData) do
			local okay, result = pcall(run, lib.serialize[test], data)
			lib.results[test] = { serialize = result }
			coroutine.yield()

			okay, result = pcall(run, lib.deserialize[test], result.data)
			lib.results[test].deserialize = result
			coroutine.yield()
		end
	end
end

function love.load()
	loadBenchmarks()
	benchmark = coroutine.create(runBenchmarks)
end

function love.update()
	local x, y = 10, 1
	frame = frame + 1
	if frame < 5 then return end
	if coroutine.status(benchmark) ~= 'dead' then
		coroutine.resume(benchmark)
	end
end

local lineHeight = math.floor(gfx.getFont():getHeight() * 1.25)

function love.keypressed(key)
	if key == 'escape' then love.event.quit() end
end

local function drawResult(lib, result, max, x, y)
	local w, h = gfx.getDimensions()
	w, h = w - 40, lineHeight

	if result.error then
		gfx.setColor(1, .4, .4)
		gfx.print(("%s FAILED: %s"):format(lib, result.error), x, y)
	else
		if result.count == max then
			gfx.setColor(.2, .6, .2)
		else
			gfx.setColor(.2, .2, 1)
		end
		gfx.rectangle('fill', x, y, w * (result.count / max), lineHeight - 2)
		gfx.setColor(1, 1, 1)
		gfx.print(("%s: %.2f ops/sec"):format(lib, result.count / result.time), x + 10, y)
	end
end

local maxCount = {
	serialize = {},
	deserialize = {},
}

function love.draw()
	gfx.clear(.2, .2, .2)

	local x, y = 10, 5
	for testName, _ in pairs(testData) do
		maxCount.serialize[testName] = maxCount.serialize[testName] or 0
		maxCount.deserialize[testName] = maxCount.deserialize[testName] or 0
		for _, part in ipairs({ 'serialize', 'deserialize' }) do
			for libName, lib in pairs(benchmarks.libraries) do
				local count = maxCount[part]
				local results = lib.results[testName]
				if results and results[part] then
					count[testName] = math.max(results[part].count, count[testName])
				end
			end
		end
	end

	for testName, _ in pairs(testData) do
		for _, part in ipairs({ 'serialize', 'deserialize' }) do
			gfx.setColor(1, 1, 1)
			gfx.print(testName .. '.' .. part, x, y)
			y = y + lineHeight
			for libName, lib in pairs(benchmarks.libraries) do
				local results = lib.results[testName]
				if results and results[part] then
					drawResult(lib.description, results[part], maxCount[part][testName], x, y)
				else
					gfx.setColor(1, 1, 1)
					gfx.print(("waiting for %s..."):format(lib.description), x, y)
				end
				y = y + lineHeight
			end
			y = y + lineHeight
		end
	end
end
